#!/usr/bin/python
#
# Copyright (C) 2016
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
import os
import logging
from optparse import OptionParser

from fabric.exceptions import NetworkError

try:
    from fabric import tasks
    from fabric.api import env, run
except ImportError as e:
    print str(e)
    print 'Please install python package fabric: sudo apt-get install fabric'
    exit(1)

from lofar.messaging import Service
from lofar.messaging import setQpidLogLevel
from lofar.common.util import waitForInterrupt
from lofar.messaging.Service import MessageHandlerInterface
import lofar.mac.config as config

logger = logging.getLogger(__name__)


class ObservationControlHandler(MessageHandlerInterface):
    def __init__(self, **kwargs):
        super(ObservationControlHandler, self).__init__(**kwargs)

        self.service2MethodMap = {
            'AbortObservation': self.abort_observation
        }

        env.hosts = ["localhost"]

        if os.environ.has_key("LOFARENV"):
            lofar_environment = os.environ['LOFARENV']

            if lofar_environment == "PRODUCTION":
                env.hosts = [config.PRODUCTION_OBSERVATION_CONTROL_HOST]
            elif lofar_environment == "TEST":
                env.hosts = [config.TEST_OBSERVATION_CONTROL_HOST]

    def _abort_observation_task(self, sas_id):
        logger.info("trying to abort ObservationControl for SAS ID: %s", sas_id)

        killed = False

        pid_line = run('pidof ObservationControl')
        pids = pid_line.split(' ')

        for pid in pids:
            pid_sas_id = run("ps -p %s --no-heading -o command | awk -F[{}] '{ printf $2; }'" % pid)
            if str(pid_sas_id) == str(sas_id):
                logger.info("Killing ObservationControl with PID: %s for SAS ID: %s", pid, sas_id)
                run('kill -SIGINT %s' % pid)
                killed = True

        return killed

    def abort_observation(self, sas_id):
        """ aborts an observation for a single sas_id """
        try:
            result = tasks.execute(self._abort_observation_task, sas_id)
            aborted = True in result.values()
        except NetworkError:
            aborted = False

        return {'aborted': aborted}

    def handle_message(self, msg):
        pass


def create_service(bus_name=config.DEFAULT_OBSERVATION_CONTROL_BUS_NAME,
                   service_name=config.DEFAULT_OBSERVATION_CONTROL_SERVICE_NAME,
                   broker=None, verbose=False):
    return Service(service_name,
                   ObservationControlHandler,
                   busname=bus_name,
                   broker=broker,
                   use_service_methods=True,
                   numthreads=1,
                   handler_args={},
                   verbose=verbose)


def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    # Check the invocation arguments
    parser = OptionParser("%prog [options]",
                          description='runs the observationcontrol service')
    parser.add_option('-q', '--broker', dest='broker', type='string', default=None, help='Address of the qpid broker, default: localhost')
    parser.add_option("-b", "--busname", dest="busname", type="string", default=config.DEFAULT_OBSERVATION_CONTROL_BUS_NAME, help="Name of the bus exchange on the qpid broker, default: %s" % config.DEFAULT_OBSERVATION_CONTROL_BUS_NAME)
    parser.add_option("-s", "--servicename", dest="servicename", type="string", default=config.DEFAULT_OBSERVATION_CONTROL_SERVICE_NAME, help="Name for this service, default: %s" % config.DEFAULT_OBSERVATION_CONTROL_SERVICE_NAME)
    parser.add_option('-V', '--verbose', dest='verbose', action='store_true', help='verbose logging')
    (options, args) = parser.parse_args()

    setQpidLogLevel(logging.INFO)
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.DEBUG if options.verbose else logging.INFO)

    with create_service(bus_name=options.busname,
                        service_name=options.servicename,
                        broker=options.broker,
                        verbose=options.verbose):
        waitForInterrupt()


if __name__ == '__main__':
    main()
