import unittest
import uuid

import mock
import os

import time
from lofar.mac.ObservationControl2 import ObservationControlHandler, create_service
from qpid.messaging.message import Message as QpidMessage
import lofar.mac.config as config

class TestObservationControlHandler(unittest.TestCase):
    pid1 = "1000"
    pid2 = "2000"

    sas_id = "100"

    def _run_side_effect(self, cmd):
        if cmd.startswith("ps -p %s" % self.pid1):
            return self.sas_id
        elif cmd.startswith("ps -p %s" % self.pid2):
            return self.sas_id + "10"
        elif cmd.startswith("pidof"):
            return "%s %s" % (self.pid1, self.pid2)
        elif cmd.startswith("kill"):
            return ""

    def setUp(self):
        fabric_run_pathcher = mock.patch('lofar.mac.ObservationControl2.run')
        self.addCleanup(fabric_run_pathcher.stop)
        self.fabric_run_mock = fabric_run_pathcher.start()

        self.fabric_run_mock.side_effect = self._run_side_effect

        fabric_tasks_pathcher = mock.patch('lofar.mac.ObservationControl2.tasks')
        self.addCleanup(fabric_tasks_pathcher.stop)
        self.fabric_tasks_mock = fabric_tasks_pathcher.start()

        fabric_env_pathcher = mock.patch('lofar.mac.ObservationControl2.env')
        self.addCleanup(fabric_env_pathcher.stop)
        self.fabric_env_mock = fabric_env_pathcher.start()

        logger_patcher = mock.patch('lofar.mac.ObservationControl2.logger')
        self.addCleanup(logger_patcher.stop)
        self.logger_mock = logger_patcher.start()

        self.observation_control_handler = ObservationControlHandler()

    def test_abort_observation_task_should_run_pidof_ObservationControl(self):
        self.observation_control_handler._abort_observation_task(self.sas_id)

        self.fabric_run_mock.assert_any_call('pidof ObservationControl')

    def test_abort_observation_tasks_should_run_ps_to_find_sas_id_on_command(self):
        self.observation_control_handler._abort_observation_task(self.sas_id)

        self.fabric_run_mock.assert_any_call(
            "ps -p %s --no-heading -o command | awk -F[{}] '{ print $2; }'" % self.pid1)
        self.fabric_run_mock.assert_any_call(
            "ps -p %s --no-heading -o command | awk -F[{}] '{ print $2; }'" % self.pid2)

    def test_abort_observation_task_should_run_kill_when_sas_id_matches(self):
        self.observation_control_handler._abort_observation_task(self.sas_id)

        self.fabric_run_mock.assert_any_call('kill -SIGINT %s' % self.pid1)

    @mock.patch.dict(os.environ, {'LOFARENV': 'TEST'})
    def test_observation_control_should_select_test_host_if_lofar_environment_is_test(self):
        ObservationControlHandler()

        self.assertEqual(self.fabric_env_mock.hosts, [config.TEST_OBSERVATION_CONTROL_HOST])

    @mock.patch.dict(os.environ, {'LOFARENV': 'PRODUCTION'})
    def test_observation_control_should_select_production_host_if_lofar_environment_is_production(self):
        ObservationControlHandler()

        self.assertEqual(self.fabric_env_mock.hosts, [config.PRODUCTION_OBSERVATION_CONTROL_HOST])

    def test_observation_control_should_select_local_host_if_no_lofar_environment_is_set(self):
        ObservationControlHandler()

        self.assertEqual(self.fabric_env_mock.hosts, ["localhost"])

    def test_abort_observation_should_execute_abort_observation_task_on_localhost(self):
        self.observation_control_handler.abort_observation(self.sas_id)

        self.fabric_tasks_mock.execute.assert_any_call(self.observation_control_handler._abort_observation_task,
                                                       self.sas_id)

    def test_abort_observation_task_should_return_false_on_unknown_sas_id(self):
        self.assertFalse(self.observation_control_handler._abort_observation_task("unknown"))

    def test_abort_observation_task_should_return_true_on_known_sas_id(self):
        self.assertTrue(self.observation_control_handler._abort_observation_task(self.sas_id))

    def test_abort_observation_task_should_log_call(self):
        self.observation_control_handler._abort_observation_task(self.sas_id)

        self.logger_mock.info.assert_any_call("trying to abort ObservationControl for SAS ID: %s", self.sas_id)

    def test_abort_observation_taks_should_log_the_kill(self):
        self.observation_control_handler._abort_observation_task(self.sas_id)

        self.logger_mock.info.assert_any_call(
            "Killing ObservationControl with PID: %s for SAS ID: %s", self.pid1, self.sas_id)

    def test_abort_observation_should_return_aborted_true_if_execute_returns_true(self):
        self.fabric_tasks_mock.execute.return_value = {'localhost': True}

        result = self.observation_control_handler.abort_observation(self.sas_id)

        self.assertTrue(result['aborted'])

    def test_abort_observation_should_return_aborted_false_if_execute_returns_false(self):
        self.fabric_tasks_mock.execute.return_value = {'localhost': False}

        result = self.observation_control_handler.abort_observation(self.sas_id)

        self.assertFalse(result['aborted'])

    @mock.patch('lofar.mac.ObservationControl2.ObservationControlHandler.abort_observation')
    @mock.patch('lofar.messaging.messagebus.qpid.messaging')
    def test_service_runs_abort_observation_task_when_called_with_AbortObservation_command(self, qpid_mock, abort_observation_mock):
        message_id = str(uuid.uuid1())
        test_id = 200

        qpid_message = QpidMessage([{'sas_id': 100}],
        reply_to="localhost",
        properties={
            "has_args": True,
            "has_kwargs": True,
            "SystemName": "LOFAR",
            "MessageType": "RequestMessage",
            "MessageId": message_id,
            "status": "OK"
        },
        subject=".AbortObservation")

        sender_mock = mock.MagicMock()
        receiver_mock = mock.MagicMock()

        receiver_mock.fetch.return_value = qpid_message
        abort_observation_mock.return_value = {"aborted": True}

        qpid_mock.Message = QpidMessage
        qpid_mock.Connection().session().senders = [sender_mock]
        qpid_mock.Connection().session().next_receiver.return_value = receiver_mock

        with create_service() as service:
            while receiver_mock.fetch.calles == 0:
                pass

        self.assertTrue(abort_observation_mock.called)


if __name__ == "__main__":
    unittest.main()