#!/usr/bin/env python

import sys

from lofar.mac.PipelineControl import *
from lofar.sas.otdb.OTDBBusListener import OTDBBusListener
from lofar.sas.otdb.config import DEFAULT_OTDB_NOTIFICATION_SUBJECT, DEFAULT_OTDB_SERVICENAME
from lofar.sas.resourceassignment.resourceassignmentservice.config import DEFAULT_SERVICENAME as DEFAULT_RAS_SERVICENAME
from lofar.messaging import ToBus, Service, EventMessage, MessageHandlerInterface

from lofar.common.methodtrigger import MethodTrigger

import subprocess
import unittest
import uuid
import datetime

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)

try:
  from mock import patch
except ImportError:
  print "Cannot run test without python MagicMock"
  print "Call 'pip install mock' / 'apt-get install python-mock'"
  exit(3)

def setUpModule():
  pass

def tearDownModule():
  pass

class TestRunCommand(unittest.TestCase):
  def test_basic(self):
    """ Test whether we can run a trivial command. """
    runCommand("true")

  def test_invalid_command(self):
    """ Test whether an invalid command produces an error. """
    with self.assertRaises(subprocess.CalledProcessError):
      output = runCommand("invalid-command")

  def test_shell(self):
    """ Test whether the command is parsed by a shell. """
    runCommand("true --version")

  def test_output(self):
    """ Test whether we catch the command output correctly. """
    output = runCommand("echo yes")
    self.assertEqual(output, "yes")

  def test_input(self):
    """ Test whether we can provide input. """
    output = runCommand("cat -", "yes")
    self.assertEqual(output, "yes")

class TestPipelineControlClassMethods(unittest.TestCase):
  def test_shouldHandle(self):
    """ Test whether we filter the right OTDB trees. """

    logger.warn('TEST_SHOULDHANDLE')

    trials = [ { "type": "Observation", "cluster": "CEP2", "shouldHandle": False },
               { "type": "Observation", "cluster": "CEP4", "shouldHandle": False },
               { "type": "Observation", "cluster": "foo",  "shouldHandle": False },
               { "type": "Observation", "cluster": "",     "shouldHandle": False },
               { "type": "Pipeline",    "cluster": "CEP2", "shouldHandle": False },
               { "type": "Pipeline",    "cluster": "CEP4", "shouldHandle": True },
               { "type": "Pipeline",    "cluster": "foo",  "shouldHandle": True },
               { "type": "Pipeline",    "cluster": "",     "shouldHandle": False },
             ]

    for t in trials:
      parset = { "ObsSW.Observation.processType": t["type"],
                 "ObsSW.Observation.Cluster.ProcessingCluster.clusterName": t["cluster"] }
      self.assertEqual(PipelineControl._shouldHandle(Parset(parset)), t["shouldHandle"])

    logger.warn('END TEST_SHOULDHANDLE')

class MockRAService(MessageHandlerInterface):
  """
    Fakes RAService calls.

    For each job, radb_id = otdb_id + 1000 to detect misplaced ids.
  """
  def __init__(self, predecessors, status):
    super(MockRAService, self).__init__()

    self.service2MethodMap = {
      "GetTask":  self.GetTask,
      "GetTasks": self.GetTasks,
    }

    self.predecessors = predecessors
    self.successors   = {x: [s for s in predecessors if x in predecessors[s]] for x in predecessors}
    self.status       = status

  def GetTask(self, id, mom_id, otdb_id, **kwargs):
    logger.info("***** GetTask(%s) *****", otdb_id)

    return {
      'status': self.status[otdb_id],

      'predecessor_ids': [1000 + x for x in self.predecessors[otdb_id]],
      'successor_ids':   [1000 + x for x in self.successors[otdb_id]],

      'starttime': datetime.datetime.utcnow(),
      'endtime': datetime.datetime.utcnow(),
    }

  def GetTasks(self, lower_bound, upper_bound, task_ids, task_status, task_type, mom_ids=None, otdb_ids=None, **kwargs):
    logger.info("***** GetTasks(%s) *****", task_ids)

    if task_ids is None:
      # Used on startup to check which tasks are at scheduled
      return []

    return [{
      'otdb_id': t - 1000,
      'status': self.status[t - 1000],

      'starttime': datetime.datetime.utcnow(),
      'endtime': datetime.datetime.utcnow(),
    } for t in task_ids]


# ================================
# Global state to manipulate
# ================================

otdb_predecessors = {}
otdb_status = {}

class MockOTDBService(MessageHandlerInterface):
    def __init__(self, notification_bus):
        """
            notification_bus: bus to send state changes to
        """
        super(MockOTDBService, self).__init__()

        self.service2MethodMap = {
            "TaskGetSpecification": self.TaskGetSpecification,
            "TaskSetStatus": self.TaskSetStatus,
            "TaskGetStatus": self.TaskGetStatus,
        }

        self.notification_bus = notification_bus

    def TaskGetSpecification(self, OtdbID):
        logger.info("***** TaskGetSpecification(%s) *****", OtdbID)

        return { "TaskSpecification": {
            "Version.number":                                                    "1",
            PARSET_PREFIX + "Observation.otdbID":                                str(OtdbID),
            PARSET_PREFIX + "Observation.processType":                           ("Observation" if OtdbID == 4 else "Pipeline"),
            PARSET_PREFIX + "Observation.Cluster.ProcessingCluster.clusterName":          "CEP4",
            PARSET_PREFIX + "Observation.Cluster.ProcessingCluster.clusterPartition":     "cpu",
            PARSET_PREFIX + "Observation.Cluster.ProcessingCluster.numberOfCoresPerTask": "2",
            PARSET_PREFIX + "Observation.Cluster.ProcessingCluster.numberOfTasks":        "110",
        } }

    def TaskSetStatus(self, OtdbID, NewStatus, UpdateTimestamps):
        logger.info("***** TaskSetStatus(%s,%s) *****", OtdbID, NewStatus)

        global otdb_status
        otdb_status[OtdbID] = NewStatus

        # Broadcast the state change
        content = { "treeID" : OtdbID, "state" : NewStatus, "time_of_change" : datetime.datetime.utcnow() }
        msg = EventMessage(context=DEFAULT_OTDB_NOTIFICATION_SUBJECT, content=content)
        self.notification_bus.send(msg)

        return {'OtdbID':OtdbID, 'MomID':None, 'Success':True}

    def TaskGetStatus(self, otdb_id, **kwargs):
        logger.info("***** TaskGetStatus(%s) *****", otdb_id)
        return {'OtdbID':otdb_id, 'status': otdb_status.get(otdb_id, 'unknown')}


class TestPipelineDependencies(unittest.TestCase):
  def setUp(self):
    # Create a random bus
    self.busname = "%s-%s" % (sys.argv[0], str(uuid.uuid4())[:8])
    self.bus = ToBus(self.busname, { "create": "always", "delete": "always", "node": { "type": "topic" } })
    self.bus.open()
    self.addCleanup(self.bus.close)

    # ================================
    # Global state to manipulate
    # ================================

    global otdb_predecessors
    otdb_predecessors = {
      1: [2,3,4],
      2: [3],
      3: [],
      4: [],
    }

    global otdb_status
    otdb_status = {
      1: "scheduled", # cannot start, since predecessor 2 hasn't finished
      2: "scheduled", # can start, since predecessor 3 has finished
      3: "finished",
      4: "scheduled", # can start, because no predecessors
    }

    # Setup mock otdb service
    service = Service(DEFAULT_OTDB_SERVICENAME,
                      MockOTDBService,
                      busname=self.busname,
                      use_service_methods=True,
                      handler_args={ "notification_bus": self.bus })
    service.start_listening()
    self.addCleanup(service.stop_listening)

    # ================================
    # Setup mock ra service
    #
    # Note that RA IDs are the same as
    # OTDB IDs + 1000 in this test.
    # ================================

    service = Service(DEFAULT_RAS_SERVICENAME,
                      MockRAService,
                      busname=self.busname,
                      use_service_methods=True,
                      handler_args={"predecessors": otdb_predecessors, "status": otdb_status})
    service.start_listening()
    self.addCleanup(service.stop_listening)

  def testGetState(self):
    with PipelineDependencies(ra_service_busname=self.busname, otdb_service_busname=self.busname) as pipelineDependencies:
      self.assertEqual(pipelineDependencies.getState(1), "scheduled")
      self.assertEqual(pipelineDependencies.getState(2), "scheduled")
      self.assertEqual(pipelineDependencies.getState(3), "finished")
      self.assertEqual(pipelineDependencies.getState(4), "scheduled")

  def testPredecessorStates(self):
    with PipelineDependencies(ra_service_busname=self.busname, otdb_service_busname=self.busname) as pipelineDependencies:
      self.assertEqual(pipelineDependencies.getPredecessorStates(1), {2: "scheduled", 3: "finished", 4: "scheduled"})
      self.assertEqual(pipelineDependencies.getPredecessorStates(3), {})

  def testSuccessorIds(self):
    with PipelineDependencies(ra_service_busname=self.busname, otdb_service_busname=self.busname) as pipelineDependencies:
      self.assertEqual(pipelineDependencies.getSuccessorIds(1), [])
      self.assertEqual(pipelineDependencies.getSuccessorIds(3), [1,2])

  def testCanStart(self):
    with PipelineDependencies(ra_service_busname=self.busname, otdb_service_busname=self.busname) as pipelineDependencies:
      self.assertEqual(pipelineDependencies.canStart(1), False)
      self.assertEqual(pipelineDependencies.canStart(2), True)
      self.assertEqual(pipelineDependencies.canStart(3), False)
      self.assertEqual(pipelineDependencies.canStart(4), True)

class TestPipelineControl(unittest.TestCase):
  def setUp(self):
    # Create a random bus
    self.busname = "%s-%s" % (sys.argv[0], str(uuid.uuid4())[:8])
    self.bus = ToBus(self.busname, { "create": "always", "delete": "always", "node": { "type": "topic" } })
    self.bus.open()
    self.addCleanup(self.bus.close)

    # Patch SLURM
    class MockSlurm(object):
      def __init__(self, *args, **kwargs):
        self.scheduled_jobs = {}

      def submit(self, jobName, *args, **kwargs):
        logger.info("Schedule SLURM job '%s': %s, %s", jobName, args, kwargs)

        self.scheduled_jobs[jobName] = (args, kwargs)

        # Return fake job ID
        return "42"

      def isQueuedOrRunning(self, otdbId):
        return str(otdbId) in self.scheduled_jobs

    patcher = patch('lofar.mac.PipelineControl.Slurm')
    patcher.start().side_effect = MockSlurm
    self.addCleanup(patcher.stop)

    # Catch functions to prevent running executables
    patcher = patch('lofar.mac.PipelineControl.Parset.dockerImage')
    patcher.start().return_value = "lofar-pipeline:trunk"
    self.addCleanup(patcher.stop)

    # ================================
    # Global state to manipulate
    # ================================

    global otdb_predecessors
    otdb_predecessors = {
      1: [2,3,4],
      2: [3],
      3: [],
      4: [],
    }

    global otdb_status
    otdb_status = {
      1: "prescheduled",
      2: "prescheduled",
      3: "prescheduled",
      4: "prescheduled",
    }

    service = Service(DEFAULT_OTDB_SERVICENAME,
                      MockOTDBService,
                      busname=self.busname,
                      use_service_methods=True,
                      handler_args={ "notification_bus": self.bus })
    service.start_listening()
    self.addCleanup(service.stop_listening)

    # ================================
    # Setup mock ra service
    # ================================

    service = Service(DEFAULT_RAS_SERVICENAME,
                      MockRAService,
                      busname=self.busname,
                      use_service_methods=True,
                      handler_args={"predecessors": otdb_predecessors, "status": otdb_status})
    service.start_listening()
    self.addCleanup(service.stop_listening)

    # ================================
    # Setup listener to catch result
    # of our service
    # ================================

    listener = OTDBBusListener(busname=self.busname)
    listener.start_listening()
    self.addCleanup(listener.stop_listening)

    self.queued_trigger = MethodTrigger(listener, "onObservationQueued")

  def test_setStatus(self):
    with PipelineControl(otdb_notification_busname=self.busname, otdb_service_busname=self.busname, ra_service_busname=self.busname) as pipelineControl:
      pipelineControl._setStatus(12345, "queued")

      # Wait for the status to propagate
      self.assertTrue(self.queued_trigger.wait())
      self.assertEqual(self.queued_trigger.args[0], 12345)

  def testNoPredecessors(self):
    """
      Request to start a simulated obsid 3, with the following predecessor tree:

        3 requires nothing
    """
    with PipelineControl(otdb_notification_busname=self.busname, otdb_service_busname=self.busname, ra_service_busname=self.busname) as pipelineControl:
      # Send fake status update
      pipelineControl._setStatus(3, "scheduled")

      # Wait for pipeline to be queued
      self.assertTrue(self.queued_trigger.wait())

      # Verify message
      self.assertEqual(self.queued_trigger.args[0], 3) # otdbId

      # Check if job was scheduled
      self.assertIn("3", pipelineControl.slurm.scheduled_jobs)
      self.assertIn("3-abort-trigger", pipelineControl.slurm.scheduled_jobs)

  def testPredecessors(self):
    """
      Request to start a simulated obsid 1, with the following predecessor tree:

        1 requires 2, 3, 4
        2 requires 3
        4 is an observation
    """
    with PipelineControl(otdb_notification_busname=self.busname, otdb_service_busname=self.busname, ra_service_busname=self.busname) as pipelineControl:
      # Send fake status update
      pipelineControl._setStatus(1, "scheduled")

      # Message should not arrive, as predecessors havent finished
      self.assertFalse(self.queued_trigger.wait())

      # Finish predecessors
      pipelineControl._setStatus(2, "finished")
      pipelineControl._setStatus(3, "finished")
      pipelineControl._setStatus(4, "finished")

      # Wait for pipeline to be queued
      self.assertTrue(self.queued_trigger.wait())

      # Verify message
      self.assertEqual(self.queued_trigger.args[0], 1) # otdbId

      # Check if job was scheduled
      self.assertIn("1", pipelineControl.slurm.scheduled_jobs)
      self.assertIn("1-abort-trigger", pipelineControl.slurm.scheduled_jobs)

def main(argv):
  unittest.main()

if __name__ == "__main__":
  # run all tests
  import sys
  main(sys.argv[1:])

